#!/bin/sh
# This script is used to generate kernels.h.

cat << EOF
// AUTOMATICALLY GENERATED FILE
//
// This file is automatically generated using kernels.h.sh
// to facilitate adding new models to the code.
// Edit the code in the kernels.h.sh instead.

static const char* kernels_source = R"source(
EOF

cat types.h states.h geometry.h models.hpp

cat << EOF
)source";
EOF
